@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2024 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
+
 #include "mupdf/fitz.h"
 
 #define SUBSCRIPT_OFFSET 0.2f
@@ -6,6 +28,52 @@
 #include <ft2build.h>
 #include FT_FREETYPE_H
 
+// Text black color when converted from DeviceCMYK to RGB
+#define CMYK_BLACK 0x221f1f
+
+static void
+scale_run(fz_context *ctx, fz_stext_block *block, float scale)
+{
+	fz_matrix m = fz_scale(scale, scale);
+	fz_stext_line *line;
+	fz_stext_char *ch;
+
+	while (block)
+	{
+		block->bbox = fz_transform_rect(block->bbox, m);
+		switch (block->type)
+		{
+		case FZ_STEXT_BLOCK_TEXT:
+			for (line = block->u.t.first_line; line; line = line->next)
+			{
+				line->bbox = fz_transform_rect(block->bbox, m);
+				for (ch = line->first_char; ch; ch = ch->next)
+				{
+					ch->origin = fz_transform_point(ch->origin, m);
+					ch->quad = fz_transform_quad(ch->quad, m);
+					ch->size = ch->size * scale;
+				}
+			}
+			break;
+
+		case FZ_STEXT_BLOCK_IMAGE:
+			block->u.i.transform = fz_post_scale(block->u.i.transform, scale, scale);
+			break;
+
+		case FZ_STEXT_BLOCK_STRUCT:
+			if (block->u.s.down)
+				scale_run(ctx, block->u.s.down->first_block, scale);
+			break;
+		}
+		block = block->next;
+	}
+}
+
+static void fz_scale_stext_page(fz_context *ctx, fz_stext_page *page, float scale)
+{
+	scale_run(ctx, page->first_block, scale);
+}
+
 /* HTML output (visual formatting with preserved layout) */
 
 static int
@@ -24,10 +92,26 @@ font_full_name(fz_context *ctx, fz_font *font)
 	return s ? s + 1 : name;
 }
 
+static const char *
+html_clean_font_name(const char *fontname)
+{
+	if (strstr(fontname, "Times"))
+		return "Times New Roman";
+	if (strstr(fontname, "Arial") || strstr(fontname, "Helvetica"))
+	{
+		if (strstr(fontname, "Narrow") || strstr(fontname, "Condensed"))
+			return "Arial Narrow";
+		return "Arial";
+	}
+	if (strstr(fontname, "Courier"))
+		return "Courier";
+	return fontname;
+}
+
 static void
 font_family_name(fz_context *ctx, fz_font *font, char *buf, int size, int is_mono, int is_serif)
 {
-	const char *name = font_full_name(ctx, font);
+	const char *name = html_clean_font_name(font_full_name(ctx, font));
 	char *s;
 	fz_strlcpy(buf, name, size);
 	s = strrchr(buf, '-');
@@ -55,14 +139,14 @@ fz_print_style_begin_html(fz_context *ctx, fz_output *out, fz_font *font, float
 	if (is_mono) fz_write_string(ctx, out, "<tt>");
 	if (is_bold) fz_write_string(ctx, out, "<b>");
 	if (is_italic) fz_write_string(ctx, out, "<i>");
-	fz_write_printf(ctx, out, "<span style=\"font-family:%s;font-size:%gpt", family, size);
-	if (color != 0)
+	fz_write_printf(ctx, out, "<span style=\"font-family:%s;font-size:%.1fpt", family, size);
+	if (color != 0 && color != CMYK_BLACK)
 		fz_write_printf(ctx, out, ";color:#%06x", color);
 	fz_write_printf(ctx, out, "\">");
 }
 
 static void
-fz_print_style_end_html(fz_context *ctx, fz_output *out, fz_font *font, float size, int sup)
+fz_print_style_end_html(fz_context *ctx, fz_output *out, fz_font *font, float size, int sup, int color)
 {
 	int is_mono = fz_font_is_monospaced(ctx, font);
 	int is_bold = fz_font_is_bold(ctx,font);
@@ -78,12 +162,128 @@ fz_print_style_end_html(fz_context *ctx, fz_output *out, fz_font *font, float si
 static void
 fz_print_stext_image_as_html(fz_context *ctx, fz_output *out, fz_stext_block *block)
 {
+	fz_matrix ctm = block->u.i.transform;
+
+#define USE_CSS_MATRIX_TRANSFORMS
+#ifdef USE_CSS_MATRIX_TRANSFORMS
+	/* Matrix maths notes.
+	 * When we get here ctm maps the unit square to the position in device
+	 * space occupied by the image.
+	 *
+	 * That is to say that mapping the 4 corners of the unit square through
+	 * the transform, give us the 4 target corners. We extend the corners
+	 * by adding an extra '1' into them to allow transforms to work. Thus
+	 * (x,y) maps through ctm = (a b c d e f) as:
+	 *
+	 * (x y 1) (a b 0) = (X Y 1)
+	 *         (c d 0)
+	 *         (e f 1)
+	 *
+	 * To simplify reading of matrix maths, we use the trick where we
+	 * 'drop' the first matrix down the page. Thus the corners c0=(0,0),
+	 * c1=(1,0), c2=(1,1), c3=(0,1) map to C0, C1, C2, C3 respectively:
+	 *
+	 *         (    a     b 0)
+	 *         (    c     d 0)
+	 *         (    e     f 1)
+	 * (0 0 1) (    e     f 1)
+	 * (0 1 1) (  c+e   d+f 1)
+	 * (1 1 1) (a+c+e b+d+f 1)
+	 * (1 0 1) (  a+e   b+f 1)
+	 *
+	 * where C0 = (e,f), C1=(c+e, d+f) C2=(a+c+e, b+d+f), C3=(a+e, b+f)
+	 *
+	 * Unfortunately, the CSS matrix transform, does not map the unit square.
+	 * Rather it does something moderately mad. As far as I can work out, the
+	 * top left corner of a (0,0) -> (w, h) box is transformed using the .e
+	 * and .f entries of the matrix. Then the image from within that square
+	 * is transformed using the centre of that square as the origin.
+	 *
+	 * So, an image placed at (0,0) in destination space with 1:1 transform
+	 * will result in an image a (0,0) as you'd expect. But an image at (0,0)
+	 * with a scale of 2, will result in 25% of the image off the left of the
+	 * screen, and 25% off the top.
+	 *
+	 * Accordingly, we have to adjust the ctm in several steps.
+	 */
+	/* Move to moving the centre of the image. */
+	ctm.e += (ctm.a+ctm.c)/2;
+	ctm.f += (ctm.b+ctm.d)/2;
+	/* Move from transforming the unit square to w/h */
+	ctm.a /= block->u.i.image->w;
+	ctm.b /= block->u.i.image->w;
+	ctm.c /= block->u.i.image->h;
+	ctm.d /= block->u.i.image->h;
+	/* Move from points to pixels */
+	ctm.a *= 96.0f/72;
+	ctm.b *= 96.0f/72;
+	ctm.c *= 96.0f/72;
+	ctm.d *= 96.0f/72;
+	ctm.e *= 96.0f/72;
+	ctm.f *= 96.0f/72;
+	/* Move to moving the top left of the untransformed image box, cos HTML is bonkers. */
+	ctm.e -= block->u.i.image->w/2;
+	ctm.f -= block->u.i.image->h/2;
+
+	fz_write_printf(ctx, out, "<img style=\"position:absolute;transform:matrix(%g,%g,%g,%g,%g,%g)\" src=\"",
+		ctm.a, ctm.b, ctm.c, ctm.d, ctm.e, ctm.f);
+#else
+	/* Alternative version of the code that uses scaleX/Y and rotate
+	 * instead, but only copes with axis aligned cases. */
+	int t;
+
 	int x = block->bbox.x0;
 	int y = block->bbox.y0;
 	int w = block->bbox.x1 - block->bbox.x0;
 	int h = block->bbox.y1 - block->bbox.y0;
 
-	fz_write_printf(ctx, out, "<img style=\"position:absolute;top:%dpt;left:%dpt;width:%dpt;height:%dpt\" src=\"", y, x, w, h);
+	const char *flip = "";
+
+	if (ctm.b == 0 && ctm.c == 0)
+	{
+		if (ctm.a < 0 && ctm.d < 0)
+			flip = "transform: scaleX(-1) scaleY(-1);";
+		else if (ctm.a < 0)
+		{
+			flip = "transform: scaleX(-1);";
+		}
+		else if (ctm.d < 0)
+		{
+			flip = "transform: scaleY(-1);";
+		}
+	} else if (ctm.a == 0 && ctm.d == 0) {
+		if (ctm.b < 0 && ctm.c < 0)
+		{
+			flip = "transform: scaleY(-1) rotate(90deg);";
+			x += (w-h)/2;
+			y -= (w-h)/2;
+			t = w; w = h; h = t;
+		}
+		else if (ctm.b < 0)
+		{
+			flip = "transform: scaleX(-1) scaleY(-1) rotate(90deg);";
+			x += (w-h)/2;
+			y -= (w-h)/2;
+			t = w; w = h; h = t;
+		}
+		else if (ctm.c < 0)
+		{
+			flip = "transform: scaleX(-1) scaleY(-1) rotate(270deg);";
+			x += (w-h)/2;
+			y -= (w-h)/2;
+			t = w; w = h; h = t;
+		}
+		else
+		{
+			flip = "transform: scaleY(-1) rotate(270deg);";
+			x += (w-h)/2;
+			y -= (w-h)/2;
+			t = w; w = h; h = t;
+		}
+	}
+
+	fz_write_printf(ctx, out, "<img style=\"position:absolute;%stop:%dpt;left:%dpt;width:%dpt;height:%dpt\" src=\"", flip, y, x, w, h);
+#endif
 	fz_write_image_as_data_uri(ctx, out, block->u.i.image);
 	fz_write_string(ctx, out, "\">\n");
 }
@@ -93,31 +293,38 @@ fz_print_stext_block_as_html(fz_context *ctx, fz_output *out, fz_stext_block *bl
 {
 	fz_stext_line *line;
 	fz_stext_char *ch;
-	int x, y;
+	float x, y, h;
 
 	fz_font *font = NULL;
 	float size = 0;
 	int sup = 0;
-	int color = 0;
+	uint32_t color = 0;
 
 	for (line = block->u.t.first_line; line; line = line->next)
 	{
 		x = line->bbox.x0;
 		y = line->bbox.y0;
+		h = line->bbox.y1 - line->bbox.y0;
 
-		fz_write_printf(ctx, out, "<p style=\"position:absolute;white-space:pre;margin:0;padding:0;top:%dpt;left:%dpt\">", y, x);
+		if (line->first_char)
+		{
+			h = line->first_char->size;
+			y = line->first_char->origin.y - h * 0.8f;
+		}
+
+		fz_write_printf(ctx, out, "<p style=\"top:%.1fpt;left:%.1fpt;line-height:%.1fpt\">", y, x, h);
 		font = NULL;
 
 		for (ch = line->first_char; ch; ch = ch->next)
 		{
 			int ch_sup = detect_super_script(line, ch);
-			if (ch->font != font || ch->size != size || ch_sup != sup || ch->color != color)
+			if (ch->font != font || ch->size != size || ch_sup != sup || ch->argb != color)
 			{
 				if (font)
-					fz_print_style_end_html(ctx, out, font, size, sup);
+					fz_print_style_end_html(ctx, out, font, size, sup, color);
 				font = ch->font;
 				size = ch->size;
-				color = ch->color;
+				color = ch->argb;
 				sup = ch_sup;
 				fz_print_style_begin_html(ctx, out, font, size, sup, color);
 			}
@@ -139,7 +346,7 @@ fz_print_stext_block_as_html(fz_context *ctx, fz_output *out, fz_stext_block *bl
 		}
 
 		if (font)
-			fz_print_style_end_html(ctx, out, font, size, sup);
+			fz_print_style_end_html(ctx, out, font, size, sup, color);
 
 		fz_write_string(ctx, out, "</p>\n");
 	}
@@ -150,10 +357,10 @@ fz_print_stext_page_as_html(fz_context *ctx, fz_output *out, fz_stext_page *page
 {
 	fz_stext_block *block;
 
-	int w = page->mediabox.x1 - page->mediabox.x0;
-	int h = page->mediabox.y1 - page->mediabox.y0;
+	float w = page->mediabox.x1 - page->mediabox.x0;
+	float h = page->mediabox.y1 - page->mediabox.y0;
 
-	fz_write_printf(ctx, out, "<div id=\"page%d\" style=\"position:relative;width:%dpt;height:%dpt;background-color:white\">\n", id, w, h);
+	fz_write_printf(ctx, out, "<div id=\"page%d\" style=\"width:%.1fpt;height:%.1fpt\">\n", id, w, h);
 
 	for (block = page->first_block; block; block = block->next)
 	{
@@ -173,8 +380,9 @@ fz_print_stext_header_as_html(fz_context *ctx, fz_output *out)
 	fz_write_string(ctx, out, "<html>\n");
 	fz_write_string(ctx, out, "<head>\n");
 	fz_write_string(ctx, out, "<style>\n");
-	fz_write_string(ctx, out, "body{background-color:gray}\n");
-	fz_write_string(ctx, out, "div{margin:1em auto}\n");
+	fz_write_string(ctx, out, "body{background-color:slategray}\n");
+	fz_write_string(ctx, out, "div{position:relative;background-color:white;margin:1em auto;box-shadow:1px 1px 8px -2px black}\n");
+	fz_write_string(ctx, out, "p{position:absolute;white-space:pre;margin:0}\n");
 	fz_write_string(ctx, out, "</style>\n");
 	fz_write_string(ctx, out, "</head>\n");
 	fz_write_string(ctx, out, "<body>\n");
@@ -323,20 +531,41 @@ static void fz_print_stext_block_as_xhtml(fz_context *ctx, fz_output *out, fz_st
 	fz_write_printf(ctx, out, "</%s>\n", tag);
 }
 
+static void
+run_to_xhtml(fz_context *ctx, fz_stext_block *block, fz_output *out)
+{
+	while (block)
+	{
+		switch(block->type)
+		{
+		case FZ_STEXT_BLOCK_IMAGE:
+			fz_print_stext_image_as_xhtml(ctx, out, block);
+			break;
+		case FZ_STEXT_BLOCK_TEXT:
+			fz_print_stext_block_as_xhtml(ctx, out, block);
+			break;
+		case FZ_STEXT_BLOCK_STRUCT:
+			fz_write_printf(ctx, out, "<struct idx=\"%d\"",
+					block->u.s.index);
+			if (block->u.s.down)
+				fz_write_printf(ctx, out, " raw=\"%s\" std=\"%s\"",
+						block->u.s.down->raw, fz_structure_to_string(block->u.s.down->standard));
+			fz_write_printf(ctx, out, ">\n");
+			if (block->u.s.down)
+				run_to_xhtml(ctx, block->u.s.down->first_block, out);
+			fz_write_printf(ctx, out, "</struct>\n");
+			break;
+		}
+		block = block->next;
+	}
+}
+
 void
 fz_print_stext_page_as_xhtml(fz_context *ctx, fz_output *out, fz_stext_page *page, int id)
 {
-	fz_stext_block *block;
-
 	fz_write_printf(ctx, out, "<div id=\"page%d\">\n", id);
 
-	for (block = page->first_block; block; block = block->next)
-	{
-		if (block->type == FZ_STEXT_BLOCK_IMAGE)
-			fz_print_stext_image_as_xhtml(ctx, out, block);
-		else if (block->type == FZ_STEXT_BLOCK_TEXT)
-			fz_print_stext_block_as_xhtml(ctx, out, block);
-	}
+	run_to_xhtml(ctx, page->first_block, out);
 
 	fz_write_string(ctx, out, "</div>\n");
 }
@@ -366,18 +595,14 @@ fz_print_stext_trailer_as_xhtml(fz_context *ctx, fz_output *out)
 
 /* Detailed XML dump of the entire structured text data */
 
-void
-fz_print_stext_page_as_xml(fz_context *ctx, fz_output *out, fz_stext_page *page, int id)
+static void
+as_xml(fz_context *ctx, fz_stext_block *block, fz_output *out)
 {
-	fz_stext_block *block;
 	fz_stext_line *line;
 	fz_stext_char *ch;
+	int i;
 
-	fz_write_printf(ctx, out, "<page id=\"page%d\" width=\"%g\" height=\"%g\">\n", id,
-		page->mediabox.x1 - page->mediabox.x0,
-		page->mediabox.y1 - page->mediabox.y0);
-
-	for (block = page->first_block; block; block = block->next)
+	while (block)
 	{
 		switch (block->type)
 		{
@@ -406,13 +631,16 @@ fz_print_stext_page_as_xml(fz_context *ctx, fz_output *out, fz_stext_page *page,
 						name = font_full_name(ctx, font);
 						fz_write_printf(ctx, out, "<font name=\"%s\" size=\"%g\">\n", name, size);
 					}
-					fz_write_printf(ctx, out, "<char quad=\"%g %g %g %g %g %g %g %g\" x=\"%g\" y=\"%g\" color=\"#%06x\" c=\"",
+					fz_write_printf(ctx, out, "<char quad=\"%g %g %g %g %g %g %g %g\" x=\"%g\" y=\"%g\" bidi=\"%d\" color=\"#%06x\" alpha=\"#%02x\" flags=\"%d\" c=\"",
 							ch->quad.ul.x, ch->quad.ul.y,
 							ch->quad.ur.x, ch->quad.ur.y,
 							ch->quad.ll.x, ch->quad.ll.y,
 							ch->quad.lr.x, ch->quad.lr.y,
 							ch->origin.x, ch->origin.y,
-							ch->color);
+							ch->bidi,
+							ch->argb & 0xFFFFFF,
+							ch->argb>>24,
+							ch->flags);
 					switch (ch->c)
 					{
 					case '<': fz_write_string(ctx, out, "&lt;"); break;
@@ -442,26 +670,71 @@ fz_print_stext_page_as_xml(fz_context *ctx, fz_output *out, fz_stext_page *page,
 			fz_write_printf(ctx, out, "<image bbox=\"%g %g %g %g\" />\n",
 					block->bbox.x0, block->bbox.y0, block->bbox.x1, block->bbox.y1);
 			break;
+
+		case FZ_STEXT_BLOCK_STRUCT:
+			fz_write_printf(ctx, out, "<struct idx=\"%d\"", block->u.s.index);
+			if (block->u.s.down)
+				fz_write_printf(ctx, out, " raw=\"%s\" std=\"%s\"",
+						block->u.s.down->raw, fz_structure_to_string(block->u.s.down->standard));
+			fz_write_printf(ctx, out, ">\n");
+			if (block->u.s.down)
+				as_xml(ctx, block->u.s.down->first_block, out);
+			fz_write_printf(ctx, out, "</struct>\n");
+			break;
+
+		case FZ_STEXT_BLOCK_VECTOR:
+			fz_write_printf(ctx, out, "<vector bbox=\"%g %g %g %g\" stroke=\"%d\" argba=\"%08x\"/>\n",
+					block->bbox.x0, block->bbox.y0, block->bbox.x1, block->bbox.y1,
+					!!block->u.v.stroked, block->u.v.argb);
+			break;
+
+		case FZ_STEXT_BLOCK_GRID:
+			fz_write_printf(ctx, out, "<grid xpos=\"");
+			for (i = 0; i < block->u.b.xs->len; i++)
+				fz_write_printf(ctx, out, "%g ", block->u.b.xs->list[i].pos);
+			fz_write_printf(ctx, out, "\" xuncertainty=\"");
+			for (i = 0; i < block->u.b.xs->len; i++)
+				fz_write_printf(ctx, out, "%d ", block->u.b.xs->list[i].uncertainty);
+			fz_write_printf(ctx, out, "\" xmaxuncertainty=\"%d\" ypos=\"", block->u.b.xs->max_uncertainty);
+			for (i = 0; i < block->u.b.ys->len; i++)
+				fz_write_printf(ctx, out, "%g ", block->u.b.ys->list[i].pos);
+			fz_write_printf(ctx, out, "\" yuncertainty=\"");
+			for (i = 0; i < block->u.b.ys->len; i++)
+				fz_write_printf(ctx, out, "%d ", block->u.b.ys->list[i].uncertainty);
+			fz_write_printf(ctx, out, "\" ymaxuncertainty=\"%d\" />\n", block->u.b.ys->max_uncertainty);
+			break;
 		}
+		block = block->next;
 	}
+}
+
+void
+fz_print_stext_page_as_xml(fz_context *ctx, fz_output *out, fz_stext_page *page, int id)
+{
+	fz_write_printf(ctx, out, "<page id=\"page%d\" width=\"%g\" height=\"%g\">\n", id,
+		page->mediabox.x1 - page->mediabox.x0,
+		page->mediabox.y1 - page->mediabox.y0);
+
+	as_xml(ctx, page->first_block, out);
+
 	fz_write_string(ctx, out, "</page>\n");
 }
 
 /* JSON dump */
 
-void
-fz_print_stext_page_as_json(fz_context *ctx, fz_output *out, fz_stext_page *page, float scale)
+static void
+as_json(fz_context *ctx, fz_stext_block *block, fz_output *out, float scale)
 {
-	fz_stext_block *block;
 	fz_stext_line *line;
 	fz_stext_char *ch;
+	int comma = 0;
 
-	fz_write_printf(ctx, out, "{%q:[", "blocks");
-
-	for (block = page->first_block; block; block = block->next)
+	while (block)
 	{
-		if (block != page->first_block)
+		if (comma)
 			fz_write_string(ctx, out, ",");
+		comma = 1;
+
 		switch (block->type)
 		{
 		case FZ_STEXT_BLOCK_TEXT:
@@ -528,15 +801,40 @@ fz_print_stext_page_as_json(fz_context *ctx, fz_output *out, fz_stext_page *page
 			fz_write_printf(ctx, out, "%q:%d,", "w", (int)((block->bbox.x1 - block->bbox.x0) * scale));
 			fz_write_printf(ctx, out, "%q:%d}}", "h", (int)((block->bbox.y1 - block->bbox.y0) * scale));
 			break;
+
+		case FZ_STEXT_BLOCK_STRUCT:
+			fz_write_printf(ctx, out, "{%q:%q,", "type", "structure");
+			fz_write_printf(ctx, out, "%q:%d", "index", block->u.s.index);
+			if (block->u.s.down)
+			{
+				fz_write_printf(ctx, out, ",%q:%q", "raw", block->u.s.down->raw);
+				fz_write_printf(ctx, out, ",%q:%q", "std", fz_structure_to_string(block->u.s.down->standard));
+				fz_write_printf(ctx, out, ",%q:[", "contents");
+				as_json(ctx, block->u.s.down->first_block, out, scale);
+				fz_write_printf(ctx, out, "]");
+			}
+			fz_write_printf(ctx, out, "}");
+			break;
+
 		}
+		block = block->next;
 	}
+}
+
+void
+fz_print_stext_page_as_json(fz_context *ctx, fz_output *out, fz_stext_page *page, float scale)
+{
+	fz_write_printf(ctx, out, "{%q:[", "blocks");
+
+	as_json(ctx, page->first_block, out, scale);
+
 	fz_write_string(ctx, out, "]}");
 }
 
 /* Plain text */
 
-void
-fz_print_stext_page_as_text(fz_context *ctx, fz_output *out, fz_stext_page *page)
+static void
+do_as_text(fz_context *ctx, fz_output *out, fz_stext_block *first_block)
 {
 	fz_stext_block *block;
 	fz_stext_line *line;
@@ -544,10 +842,11 @@ fz_print_stext_page_as_text(fz_context *ctx, fz_output *out, fz_stext_page *page
 	char utf[10];
 	int i, n;
 
-	for (block = page->first_block; block; block = block->next)
+	for (block = first_block; block; block = block->next)
 	{
-		if (block->type == FZ_STEXT_BLOCK_TEXT)
+		switch (block->type)
 		{
+		case FZ_STEXT_BLOCK_TEXT:
 			for (line = block->u.t.first_line; line; line = line->next)
 			{
 				for (ch = line->first_char; ch; ch = ch->next)
@@ -559,10 +858,21 @@ fz_print_stext_page_as_text(fz_context *ctx, fz_output *out, fz_stext_page *page
 				fz_write_string(ctx, out, "\n");
 			}
 			fz_write_string(ctx, out, "\n");
+			break;
+		case FZ_STEXT_BLOCK_STRUCT:
+			if (block->u.s.down != NULL)
+				do_as_text(ctx, out, block->u.s.down->first_block);
+			break;
 		}
 	}
 }
 
+void
+fz_print_stext_page_as_text(fz_context *ctx, fz_output *out, fz_stext_page *page)
+{
+	do_as_text(ctx, out, page->first_block);
+}
+
 /* Text output writer */
 
 enum {
@@ -587,6 +897,7 @@ static fz_device *
 text_begin_page(fz_context *ctx, fz_document_writer *wri_, fz_rect mediabox)
 {
 	fz_text_writer *wri = (fz_text_writer*)wri_;
+	float s = wri->opts.scale;
 
 	if (wri->page)
 	{
@@ -596,7 +907,7 @@ text_begin_page(fz_context *ctx, fz_document_writer *wri_, fz_rect mediabox)
 
 	wri->number++;
 
-	wri->page = fz_new_stext_page(ctx, mediabox);
+	wri->page = fz_new_stext_page(ctx, fz_transform_rect(mediabox, fz_scale(s, s)));
 	return fz_new_stext_device(ctx, wri->page, &wri->opts);
 }
 
@@ -604,6 +915,9 @@ static void
 text_end_page(fz_context *ctx, fz_document_writer *wri_, fz_device *dev)
 {
 	fz_text_writer *wri = (fz_text_writer*)wri_;
+	float s = wri->opts.scale;
+
+	fz_scale_stext_page(ctx, wri->page, s);
 
 	fz_try(ctx)
 	{
@@ -673,11 +987,13 @@ text_drop_writer(fz_context *ctx, fz_document_writer *wri_)
 fz_document_writer *
 fz_new_text_writer_with_output(fz_context *ctx, const char *format, fz_output *out, const char *options)
 {
-	fz_text_writer *wri;
+	fz_text_writer *wri = NULL;
+
+	fz_var(wri);
 
-	wri = fz_new_derived_document_writer(ctx, fz_text_writer, text_begin_page, text_end_page, text_close_writer, text_drop_writer);
 	fz_try(ctx)
 	{
+		wri = fz_new_derived_document_writer(ctx, fz_text_writer, text_begin_page, text_end_page, text_close_writer, text_drop_writer);
 		fz_parse_stext_options(ctx, &wri->opts, options);
 
 		wri->format = FZ_FORMAT_TEXT;
@@ -718,6 +1034,7 @@ fz_new_text_writer_with_output(fz_context *ctx, const char *format, fz_output *o
 	}
 	fz_catch(ctx)
 	{
+		fz_drop_output(ctx, out);
 		fz_free(ctx, wri);
 		fz_rethrow(ctx);
 	}
@@ -729,13 +1046,5 @@ fz_document_writer *
 fz_new_text_writer(fz_context *ctx, const char *format, const char *path, const char *options)
 {
 	fz_output *out = fz_new_output_with_path(ctx, path ? path : "out.txt", 0);
-	fz_document_writer *wri = NULL;
-	fz_try(ctx)
-		wri = fz_new_text_writer_with_output(ctx, format, out, options);
-	fz_catch(ctx)
-	{
-		fz_drop_output(ctx, out);
-		fz_rethrow(ctx);
-	}
-	return wri;
+	return fz_new_text_writer_with_output(ctx, format, out, options);
 }
